'use strict';

/**
 * Callback for coordEach
 *
 * @callback coordEachCallback
 * @param {Array<number>} currentCoord The current coordinate being processed.
 * @param {number} coordIndex The current index of the coordinate being processed.
 * @param {number} featureIndex The current index of the Feature being processed.
 * @param {number} multiFeatureIndex The current index of the Multi-Feature being processed.
 * @param {number} geometryIndex The current index of the Geometry being processed.
 */

/**
 * Iterate over coordinates in any GeoJSON object, similar to Array.forEach()
 *
 * @name coordEach
 * @param {FeatureCollection|Feature|Geometry} geojson any GeoJSON object
 * @param {Function} callback a method that takes (currentCoord, coordIndex, featureIndex, multiFeatureIndex)
 * @param {boolean} [excludeWrapCoord=false] whether or not to include the final coordinate of LinearRings that wraps the ring in its iteration.
 * @returns {void}
 * @example
 * var features = turf.featureCollection([
 *   turf.point([26, 37], {"foo": "bar"}),
 *   turf.point([36, 53], {"hello": "world"})
 * ]);
 *
 * turf.coordEach(features, function (currentCoord, coordIndex, featureIndex, multiFeatureIndex, geometryIndex) {
 *   //=currentCoord
 *   //=coordIndex
 *   //=featureIndex
 *   //=multiFeatureIndex
 *   //=geometryIndex
 * });
 */
function coordEach(geojson, callback, excludeWrapCoord) {
  // Handles null Geometry -- Skips this GeoJSON
  if (geojson === null) return;
  var j,
    k,
    l,
    geometry,
    stopG,
    coords,
    geometryMaybeCollection,
    wrapShrink = 0,
    coordIndex = 0,
    isGeometryCollection,
    type = geojson.type,
    isFeatureCollection = type === "FeatureCollection",
    isFeature = type === "Feature",
    stop = isFeatureCollection ? geojson.features.length : 1;

  // This logic may look a little weird. The reason why it is that way
  // is because it's trying to be fast. GeoJSON supports multiple kinds
  // of objects at its root: FeatureCollection, Features, Geometries.
  // This function has the responsibility of handling all of them, and that
  // means that some of the `for` loops you see below actually just don't apply
  // to certain inputs. For instance, if you give this just a
  // Point geometry, then both loops are short-circuited and all we do
  // is gradually rename the input until it's called 'geometry'.
  //
  // This also aims to allocate as few resources as possible: just a
  // few numbers and booleans, rather than any temporary arrays as would
  // be required with the normalization approach.
  for (var featureIndex = 0; featureIndex < stop; featureIndex++) {
    geometryMaybeCollection = isFeatureCollection
      ? geojson.features[featureIndex].geometry
      : isFeature
      ? geojson.geometry
      : geojson;
    isGeometryCollection = geometryMaybeCollection
      ? geometryMaybeCollection.type === "GeometryCollection"
      : false;
    stopG = isGeometryCollection
      ? geometryMaybeCollection.geometries.length
      : 1;

    for (var geomIndex = 0; geomIndex < stopG; geomIndex++) {
      var multiFeatureIndex = 0;
      var geometryIndex = 0;
      geometry = isGeometryCollection
        ? geometryMaybeCollection.geometries[geomIndex]
        : geometryMaybeCollection;

      // Handles null Geometry -- Skips this geometry
      if (geometry === null) continue;
      coords = geometry.coordinates;
      var geomType = geometry.type;

      wrapShrink =
        excludeWrapCoord &&
        (geomType === "Polygon" || geomType === "MultiPolygon")
          ? 1
          : 0;

      switch (geomType) {
        case null:
          break;
        case "Point":
          if (
            callback(
              coords,
              coordIndex,
              featureIndex,
              multiFeatureIndex,
              geometryIndex
            ) === false
          )
            return false;
          coordIndex++;
          multiFeatureIndex++;
          break;
        case "LineString":
        case "MultiPoint":
          for (j = 0; j < coords.length; j++) {
            if (
              callback(
                coords[j],
                coordIndex,
                featureIndex,
                multiFeatureIndex,
                geometryIndex
              ) === false
            )
              return false;
            coordIndex++;
            if (geomType === "MultiPoint") multiFeatureIndex++;
          }
          if (geomType === "LineString") multiFeatureIndex++;
          break;
        case "Polygon":
        case "MultiLineString":
          for (j = 0; j < coords.length; j++) {
            for (k = 0; k < coords[j].length - wrapShrink; k++) {
              if (
                callback(
                  coords[j][k],
                  coordIndex,
                  featureIndex,
                  multiFeatureIndex,
                  geometryIndex
                ) === false
              )
                return false;
              coordIndex++;
            }
            if (geomType === "MultiLineString") multiFeatureIndex++;
            if (geomType === "Polygon") geometryIndex++;
          }
          if (geomType === "Polygon") multiFeatureIndex++;
          break;
        case "MultiPolygon":
          for (j = 0; j < coords.length; j++) {
            geometryIndex = 0;
            for (k = 0; k < coords[j].length; k++) {
              for (l = 0; l < coords[j][k].length - wrapShrink; l++) {
                if (
                  callback(
                    coords[j][k][l],
                    coordIndex,
                    featureIndex,
                    multiFeatureIndex,
                    geometryIndex
                  ) === false
                )
                  return false;
                coordIndex++;
              }
              geometryIndex++;
            }
            multiFeatureIndex++;
          }
          break;
        case "GeometryCollection":
          for (j = 0; j < geometry.geometries.length; j++)
            if (
              coordEach(geometry.geometries[j], callback, excludeWrapCoord) ===
              false
            )
              return false;
          break;
        default:
          throw new Error("Unknown Geometry Type");
      }
    }
  }
}

/**
 * Takes a set of features, calculates the bbox of all input features, and returns a bounding box.
 *
 * @name bbox
 * @param {GeoJSON} geojson any GeoJSON object
 * @returns {BBox} bbox extent in [minX, minY, maxX, maxY] order
 * @example
 * var line = turf.lineString([[-74, 40], [-78, 42], [-82, 35]]);
 * var bbox = turf.bbox(line);
 * var bboxPolygon = turf.bboxPolygon(bbox);
 *
 * //addToMap
 * var addToMap = [line, bboxPolygon]
 */
function bbox(geojson) {
    var result = [Infinity, Infinity, -Infinity, -Infinity];
    coordEach(geojson, function (coord) {
        if (result[0] > coord[0]) {
            result[0] = coord[0];
        }
        if (result[1] > coord[1]) {
            result[1] = coord[1];
        }
        if (result[2] < coord[0]) {
            result[2] = coord[0];
        }
        if (result[3] < coord[1]) {
            result[3] = coord[1];
        }
    });
    return result;
}
bbox["default"] = bbox;

// it is what it is
const id = i => i;

/**
 * Get the action function based on feature properties
 *
 * First, check for an `action` property that matches a member of actions.
 * Next, try a default action.
 * - if there's a bounding box, `fitBounds`
 * - if it's a point, a point, `flyTo`
 * - if it's a polygon (or anything else, really), find a bounding box and call `fitBounds`
 *
 * @param {Object} feature a single [geojson feature](https://tools.ietf.org/html/rfc7946#section-3.2),
 * usually the one in view
 * @param {Object} actions an object with defined actions to use. Each action is a function that takes
 * a map and the current feature in view.
 * @returns {function} with the signature `action(map, feature): void`
 */
function getAction(feature = {}, actions = {}) {
	if ("action" in feature.properties && feature.properties.action in actions) {
		return actions[feature.properties.action];
	}

	// handle nulls

	if (
		!feature.geometry || // no geometry
		!feature.geometry.coordinates || // no coordinates
		0 === feature.geometry.coordinates.length // [] coordinates
	) {
		return actions.noop;
	}

	// no defined action, get a default, starting with fitBounds
	if (feature.bbox) {
		return actions.fitBounds;
	}

	// we can fly to a point
	if (feature.geometry.type === "Point") {
		return actions.flyTo;
	}

	// everything else has bounds
	return actions.fitBounds;
}

/**
 * Extract options named in keys from properties and cast to type
 *
 * @param {Object} properties A properties object from a feature from which to extract values
 * @param {Array<string>} keys A list of keys to extract
 * @param {function} cast A function to convert keys to a different type, ie `Number`
 *
 * @returns {object} An options object, with values converted by `cast`
 */
function extractOptions(properties, keys = [], cast = id) {
	return keys.reduce((m, k) => {
		if (k in properties) {
			m[k] = cast(properties[k]);
		}
		return m;
	}, {});
}

/**
 * flyTo wraps Mapbox's Map#flyTo method, extracting arguments from feature properties
 * and centering on this feature's geometry (which should be a point).
 *
 * The following properties are extracted, converted to numbers
 * and passed to [Map#flyTo](https://docs.mapbox.com/mapbox-gl-js/api/#map#flyto):
 *
 * - `zoom`
 * - `bearing`
 * - `pitch`
 * - `duration`
 *
 * @export
 * @param {mapboxgl.Map} map The map attached to our current story
 * @param {Object} feature The feature in view
 */
function flyTo(map, { geometry, properties }) {
	const options = extractOptions(
		properties,
		["zoom", "bearing", "pitch", "duration"],
		Number
	);

	if (!options.zoom) {
		options.zoom = map.getMaxZoom();
	}

	options.center = geometry.coordinates;

	map.flyTo(options);
}

/**
 * fitBounds finds a bounding box for a feature and fits the map to it,
 * using [Map#fitBounds](https://docs.mapbox.com/mapbox-gl-js/api/#map#fitbounds).
 * If the feature has a bbox attribute, that is used. Otherwise, we calculate a bounding box
 * and use that.
 *
 * The following properties are extracted, converted to numbers
 * and passed to [Map#fitBounds](https://docs.mapbox.com/mapbox-gl-js/api/#map#fitbounds):
 *
 * - `maxZoom`
 * - `bearing`
 * - `pitch`
 * - `duration`
 * - `padding` (single padding only)
 *
 * @export
 * @param {mapboxgl.Map} map The map attached to our current story
 * @param {Object} feature The feature in view
 */
function fitBounds(map, feature) {
	const options = extractOptions(
		feature.properties,
		["maxZoom", "bearing", "pitch", "duration", "padding"],
		Number
	);

	const bounds = feature.bbox ? feature.bbox : bbox(feature);

	map.fitBounds(bounds, options);
}

/**
 * noop is the best action because it wants nothing and gives nothing in return
 */
function noop() {}

var defaultActions = /*#__PURE__*/Object.freeze({
  __proto__: null,
  fitBounds: fitBounds,
  flyTo: flyTo,
  noop: noop
});

/**
 * Ulysses creates a narrative around geography by linking narrative steps to actions on a map.
 * Each step is a feature in a [GeoJSON feature collection](https://tools.ietf.org/html/rfc7946#section-3.3),
 * with a specific action defined as a property.
 *
 * Actions are functions that change the map state -- panning, zooming, rotating, adding or removing layers --
 * anything you can do with the underlying map.
 *
 * @param {mapboxgl.Map} map The map we're controlling. For now, only mapbox is supported.
 * @param {Object|Array} steps A [GeoJSON feature collection](https://tools.ietf.org/html/rfc7946#section-3.3),
 * or an array of features, where each feature is a step in our story.
 * @param {Object} actions An object containing action functions to be called for each feature.
 * This gets merged into default actions (`flyTo`, `fitBounds`, `noop`) so you can override those
 * defaults by adding new functions of the same name.
 * @param {Array[Function]} plugins An array of plugin functions to apply to this instance
 *
 * @class Ulysses
 */
class Ulysses {
	constructor({ map, steps = {}, actions = {}, plugins = [] }) {
		this.map = map;
		if (Array.isArray(steps)) {
			steps = { type: "FeatureCollection", features: steps };
		}
		this.steps = steps;
		this.actions = Object.assign({}, defaultActions, actions);

		this._current = -1;

		// define our default events
		this._callbacks = Object.assign(Object.create(null), {
			step: [],
			next: [],
			previous: [],
			start: [],
			end: [],
			destroy: [],
		});

		this._plugins = plugins.map(plugin => plugin(this));
	}

	/**
	 * Get the current step number (zero-indexed)
	 *
	 * @returns {Number}
	 * @readonly
	 * @instance Ulysses
	 * @property
	 */
	get current() {
		return this._current;
	}
	/**
	 * Get the number fo steps in this story
	 *
	 * @returns {Number}
	 * @readonly
	 * @instance Ulysses
	 */
	get length() {
		return this.steps.features.length;
	}
	/**
	 * Advance one step (unless we're at the end)
	 *
	 * @instance Ulysses
	 * @method
	 */
	next() {
		this.trigger("next", {
			step: this.current,
			feature: this.current > -1 ? this.steps.features[this.current] : undefined,
		});
		this.step(this.current + 1);
	}
	/**
	 * Go back a step (unless we're at the beginning)
	 *
	 * @instance Ulysses
	 */
	previous() {
		this.trigger("previous", {
			step: this.current,
			feature: this.steps.features[this.current],
		});
		this.step(this.current - 1);
	}
	/**
	 * Go to step `n`, where `n` is a step number
	 *
	 * @param {Number} n Go to this step
	 * @instance Ulysses
	 */
	step(n) {
		const feature = this.steps.features[n];
		if (feature === undefined) return;

		if (this.current === -1) {
			this.trigger("start");
		}
		this._current = n;

		const action = getAction(feature, this.actions);

		action(this.map, feature);

		this.trigger("step", { step: n, feature });

		if (n === this.length - 1) {
			this.trigger("end");
		}
	}

	// events
	/**
	 * Register a callback for a given event
	 *
	 * @param {String} event
	 * @param {Function} callback
	 * @instance Ulysses
	 * @returns {Function} unsubscribe function
	 */
	on(event, callback) {
		if (!Array.isArray(this._callbacks[event])) {
			this._callbacks[event] = [];
		}

		this._callbacks[event].push(callback);

		return () => {
			this.off(event, callback);
		};
	}

	/**
	 * Register a callback for an event that only fires once
	 *
	 * @param {String} event
	 * @param {Function} callback
	 * @instance Ulysses
	 */

	once(event, callback) {
		this.on(event, e => {
			this.off(event, callback);
			callback.call(this, e);
		});
	}

	/**
	 * Remove a callback for a given event. If callback is `undefined`, remove all callbacks for that event.
	 *
	 * @param {String} event
	 * @param {Function} callback
	 * @instance Ulysses
	 */
	off(event, callback) {
		// noop
		if (!Array.isArray(this._callbacks[event])) return;

		// remove all
		if (!callback) {
			this._callbacks[event] = [];
		}

		// remove one
		const i = this._callbacks[event].indexOf(callback);
		if (i > -1) {
			this._callbacks[event].splice(i, 1);
		}
	}

	/**
	 * Fire callbacks for a given event. This is used internally but may be useful in actions and plugins.
	 * The `detail` object is passed to callback functions.
	 *
	 * @param {String} event
	 * @param {Object} [detail={}]
	 * @instance Ulysses
	 */
	trigger(event, detail = {}) {
		const callbacks = this._callbacks[event];

		if (!Array.isArray(callbacks)) return;

		callbacks.forEach(f => {
			f.call(this, detail);
		});
	}

	// plugins
	/**
	 * Apply a plugin function to this Ulysses instance
	 *
	 * @param {Function} plugin
	 * @instance Ulysses
	 * @returns {ThisType}
	 */
	use(plugin) {
		this._plugins.push(plugin(this));
	}
	/**
	 * Remove event listers and call plugin cleanup functions.
	 * Triggers a "destroy" event with no data.
	 *
	 * @instance Ulysses
	 */
	destroy() {
		this.trigger("destroy");
		// remove all events
		this._callbacks = {};

		// run plugin cleanup
		this._plugins.forEach(f => {
			if (typeof f === "function") {
				f();
			}
		});
	}
}

module.exports = Ulysses;
